# SPDX-FileCopyrightText: 2021 Mark Komus for Adafruit Industries
#
# SPDX-License-Identifier: MIT

# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""
`adafruit_midi.control_change_values`
================================================================================

Definition for standard MIDI control change values.


* Author(s): Mark Komus

Implementation Notes
--------------------

"""

__version__ = "0.0.0-auto.0"
__repo__ = "https://github.com/adafruit/Adafruit_CircuitPython_MIDI.git"

MOD_WHEEL = 1
BREATH_CONTROL = 2
FOOT_CONTROLLER = 4
VOLUME = 7
PAN = 10
EXPRESSION = 11
PORTAMENTO_TIME = 5
SUSTAIN_PEDAL = 64
PORTAMENTO = 65
FILTER_RESONANCE = 71
RELEASE_TIME = 72
ATTACK_TIME = 73
CUTOFF_FREQUENCY = 74
DECAY_TIME = 75
VIBRATO_RATE = 76
VIBRATO_DEPTH = 77
VIBRATO_DELAY = 78
CHORUS = 93
ALL_CONTROLLERS_OFF = 121
ALL_NOTES_OFF = 123



# SPDX-FileCopyrightText: 2019 Kevin J. Walters for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""
`adafruit_midi.system_exclusive`
================================================================================

System Exclusive MIDI message.


* Author(s): Kevin J. Walters

Implementation Notes
--------------------

"""

from .midi_message import MIDIMessage

__version__ = "0.0.0-auto.0"
__repo__ = "https://github.com/adafruit/Adafruit_CircuitPython_MIDI.git"


class SystemExclusive(MIDIMessage):
    """System Exclusive MIDI message.

    :param list manufacturer_id: The single byte or three byte
        manufacturer's id as a list or bytearray of numbers between 0-127.
    :param list data: The 7bit data as a list or bytearray of numbers between 0-127.

    This message can only be parsed if it fits within the input buffer in :class:MIDI.
    """

    _STATUS = 0xF0
    _STATUSMASK = 0xFF
    LENGTH = -1
    ENDSTATUS = 0xF7

    def __init__(self, manufacturer_id, data):
        self.manufacturer_id = bytes(manufacturer_id)
        self.data = bytes(data)
        super().__init__()

    def __bytes__(self):
        return (
            bytes([self._STATUS])
            + self.manufacturer_id
            + self.data
            + bytes([self.ENDSTATUS])
        )

    @classmethod
    def from_bytes(cls, msg_bytes):
        # -1 on second arg is to avoid the ENDSTATUS which is passed
        if msg_bytes[1] != 0:  # pylint: disable=no-else-return
            return cls(msg_bytes[1:2], msg_bytes[2:-1])
        else:
            return cls(msg_bytes[1:4], msg_bytes[4:-1])


SystemExclusive.register_message_type()



# SPDX-FileCopyrightText: 2019 Kevin J. Walters for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""
`adafruit_midi.control_change`
================================================================================

Control Change MIDI message.


* Author(s): Kevin J. Walters

Implementation Notes
--------------------

"""

from .midi_message import MIDIMessage

__version__ = "0.0.0-auto.0"
__repo__ = "https://github.com/adafruit/Adafruit_CircuitPython_MIDI.git"


class ControlChange(MIDIMessage):
    """Control Change MIDI message.

    :param int control: The control number, 0-127.
    :param int value: The 7bit value of the control, 0-127.

    """

    _STATUS = 0xB0
    _STATUSMASK = 0xF0
    LENGTH = 3

    def __init__(self, control, value, *, channel=None):
        self.control = control
        self.value = value
        super().__init__(channel=channel)
        if not 0 <= self.control <= 127 or not 0 <= self.value <= 127:
            self._raise_valueerror_oor()

    def __bytes__(self):
        return bytes(
            [self._STATUS | (self.channel & self.CHANNELMASK), self.control, self.value]
        )

    @classmethod
    def from_bytes(cls, msg_bytes):
        return cls(msg_bytes[1], msg_bytes[2], channel=msg_bytes[0] & cls.CHANNELMASK)


ControlChange.register_message_type()



# SPDX-FileCopyrightText: 2019 Kevin J. Walters for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""
`adafruit_midi.timing_clock`
================================================================================

Timing Clock MIDI message.


* Author(s): Kevin J. Walters

Implementation Notes
--------------------

"""

from .midi_message import MIDIMessage

__version__ = "0.0.0-auto.0"
__repo__ = "https://github.com/adafruit/Adafruit_CircuitPython_MIDI.git"


# Good to have this registered first as it occurs frequently when present
class TimingClock(MIDIMessage):
    """Timing Clock MIDI message.

    This occurs 24 times per quarter note when synchronization is in use.
    If this is not needed it's best to avoid this sending this high frequency
    message to a CircuitPython device to reduce the amount of message processing.
    """

    _STATUS = 0xF8
    _STATUSMASK = 0xFF
    LENGTH = 1


TimingClock.register_message_type()



# SPDX-FileCopyrightText: 2019 Kevin J. Walters for Adafruit Industries
#
# SPDX-License-Identifier: MIT

import unittest

import os

verbose = int(os.getenv("TESTVERBOSE", "2"))

# pylint: disable=wrong-import-position
# adafruit_midi had an import usb_midi
import sys

# sys.modules['usb_midi'] = MagicMock()

# Borrowing the dhalbert/tannewt technique from adafruit/Adafruit_CircuitPython_Motor
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from adafruit_midi.midi_message import note_parser

# pylint: enable=wrong-import-position


class Test_note_parser(unittest.TestCase):
    def text_int_passthru(self):
        self.assertEqual(note_parser(0), 0)
        self.assertEqual(note_parser(70), 70)
        self.assertEqual(note_parser(127), 127)

        # it does not range check so these should pass
        self.assertEqual(note_parser(-303), -303)
        self.assertEqual(note_parser(808), 808)

    def test_good_text(self):
        note_prefix = {
            "Cb": 11,
            "C": 12,
            "C#": 13,
            "Db": 13,
            "D": 14,
            "D#": 15,
            "Eb": 15,
            "E": 16,
            "Fb": 16,
            "E#": 17,
            "F": 17,
            "F#": 18,
            "Gb": 18,
            "G": 19,
            "G#": 20,
            "Ab": 20,
            "A": 21,
            "A#": 22,
            "Bb": 22,
            "B": 23,
            "B#": 24,
        }

        # test from Cb0 to B#8
        for prefix, base_value in note_prefix.items():
            for octave in range(9):
                note = prefix + str(octave)
                expected_value = base_value + octave * 12  # 12 semitones in octave
                self.assertEqual(note_parser(note), expected_value)

        # re-test with simple C4/A4 tests to catch any bugs in above
        self.assertEqual(note_parser("C4"), 60)
        self.assertEqual(note_parser("A4"), 69)

    def test_bad_text(self):

        for text_note in ["H", "H4", "asdfasdfasdf", "000", "999"]:
            with self.assertRaises(ValueError):
                note_parser(text_note)


if __name__ == "__main__":
    unittest.main(verbosity=verbose)



# SPDX-FileCopyrightText: 2019 Limor Fried for Adafruit Industries
# SPDX-FileCopyrightText: 2019 Kevin J. Walters for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""
`adafruit_midi`
================================================================================

A CircuitPython helper for encoding/decoding MIDI packets over a MIDI or UART connection.


* Author(s): Limor Fried, Kevin J. Walters

Implementation Notes
--------------------

**Hardware:**



**Software and Dependencies:**

* Adafruit CircuitPython firmware for the supported boards:
  https://github.com/adafruit/circuitpython/releases

"""

from .midi_message import MIDIMessage

__version__ = "0.0.0-auto.0"
__repo__ = "https://github.com/adafruit/Adafruit_CircuitPython_MIDI.git"


class MIDI:
    """MIDI helper class. ``midi_in`` or ``midi_out`` *must* be set or both together.

    :param midi_in: an object which implements ``read(length)``,
        set to ``usb_midi.ports[0]`` for USB MIDI, default None.
    :param midi_out: an object which implements ``write(buffer, length)``,
        set to ``usb_midi.ports[1]`` for USB MIDI, default None.
    :param in_channel: The input channel(s).
        This is used by ``receive`` to filter data.
        This can either be an ``int`` for the wire protocol channel number (0-15)
        a tuple of ``int`` to listen for multiple channels.
        Defaults to all channels.
    :param int out_channel: The wire protocol output channel number (0-15)
        used by ``send`` if no channel is specified,
        defaults to 0 (MIDI Channel 1).
    :param int in_buf_size: Maximum size of input buffer in bytes, default 30.
    :param bool debug: Debug mode, default False.

    """

    def __init__(
        self,
        midi_in=None,
        midi_out=None,
        *,
        in_channel=None,
        out_channel=0,
        in_buf_size=30,
        debug=False
    ):
        if midi_in is None and midi_out is None:
            raise ValueError("No midi_in or midi_out provided")
        self._midi_in = midi_in
        self._midi_out = midi_out
        self._in_channel = in_channel
        self.in_channel = in_channel
        self._out_channel = out_channel
        self.out_channel = out_channel
        self._debug = debug
        # This input buffer holds what has been read from midi_in
        self._in_buf = bytearray(0)
        self._in_buf_size = in_buf_size
        self._outbuf = bytearray(4)
        self._skipped_bytes = 0

    @property
    def in_channel(self):
        """The incoming MIDI channel. Must be 0-15. Correlates to MIDI channels 1-16, e.g.
        ``in_channel = 3`` will listen on MIDI channel 4.
        Can also listen on multiple channels, e.g. ``in_channel  = (0,1,2)``
        will listen on MIDI channels 1-3.
        Default is all channels."""
        return self._in_channel

    @in_channel.setter
    def in_channel(self, channel):
        if channel is None or channel == "ALL":
            self._in_channel = tuple(range(16))
        elif isinstance(channel, int) and 0 <= channel <= 15:
            self._in_channel = channel
        elif isinstance(channel, tuple) and all(0 <= c <= 15 for c in channel):
            self._in_channel = channel
        else:
            raise RuntimeError("Invalid input channel")

    @property
    def out_channel(self):
        """The outgoing MIDI channel. Must be 0-15. Correlates to MIDI channels 1-16, e.g.
        ``out_channel = 3`` will send to MIDI channel 4. Default is 0 (MIDI channel 1)."""
        return self._out_channel

    @out_channel.setter
    def out_channel(self, channel):
        if not 0 <= channel <= 15:
            raise RuntimeError("Invalid output channel")
        self._out_channel = channel

    def receive(self):
        """Read messages from MIDI port, store them in internal read buffer, then parse that data
        and return the first MIDI message (event).
        This maintains the blocking characteristics of the midi_in port.

        :returns MIDIMessage object: Returns object or None for nothing.
        """
        ### could check _midi_in is an object OR correct object OR correct interface here?
        # If the buffer here is not full then read as much as we can fit from
        # the input port
        if len(self._in_buf) < self._in_buf_size:
            bytes_in = self._midi_in.read(self._in_buf_size - len(self._in_buf))
            if bytes_in:
                if self._debug:
                    print("Receiving: ", [hex(i) for i in bytes_in])
                self._in_buf.extend(bytes_in)
                del bytes_in

        (msg, endplusone, skipped) = MIDIMessage.from_message_bytes(
            self._in_buf, self._in_channel
        )
        if endplusone != 0:
            # This is not particularly efficient as it's copying most of bytearray
            # and deleting old one
            self._in_buf = self._in_buf[endplusone:]

        self._skipped_bytes += skipped

        # msg could still be None at this point, e.g. in middle of monster SysEx
        return msg

    def send(self, msg, channel=None):
        """Sends a MIDI message.

        :param msg: Either a MIDIMessage object or a sequence (list) of MIDIMessage objects.
            The channel property will be *updated* as a side-effect of sending message(s).
        :param int channel: Channel number, if not set the ``out_channel`` will be used.

        """
        if channel is None:
            channel = self.out_channel
        if isinstance(msg, MIDIMessage):
            msg.channel = channel
            data = msg.__bytes__()  # bytes(object) does not work in uPy
        else:
            data = bytearray()
            for each_msg in msg:
                each_msg.channel = channel
                data.extend(each_msg.__bytes__())

        self._send(data, len(data))

    def _send(self, packet, num):
        if self._debug:
            print("Sending: ", [hex(i) for i in packet[:num]])
        self._midi_out.write(packet, num)



# SPDX-FileCopyrightText: 2019 Kevin J. Walters for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""
`adafruit_midi.note_off`
================================================================================

Note Off Change MIDI message.


* Author(s): Kevin J. Walters

Implementation Notes
--------------------

"""

from .midi_message import MIDIMessage, note_parser

__version__ = "0.0.0-auto.0"
__repo__ = "https://github.com/adafruit/Adafruit_CircuitPython_MIDI.git"


class NoteOff(MIDIMessage):  # pylint: disable=duplicate-code
    """Note Off Change MIDI message.

    :param note: The note (key) number either as an ``int`` (0-127) or a
        ``str`` which is parsed, e.g. "C4" (middle C) is 60, "A4" is 69.
    :param int velocity: The release velocity, 0-127, defaults to 0.

    """

    _STATUS = 0x80
    _STATUSMASK = 0xF0
    LENGTH = 3

    def __init__(self, note, velocity=0, *, channel=None):
        self.note = note_parser(note)
        """Key, either int (0-127) or string that will be turned off """
        self.velocity = velocity
        """Release velocity, int (0-127) """
        super().__init__(channel=channel)
        if not 0 <= self.note <= 127 or not 0 <= self.velocity <= 127:
            self._raise_valueerror_oor()

    def __bytes__(self):
        return bytes(
            [
                self._STATUS | (self.channel & self.CHANNELMASK),
                self.note,
                self.velocity,
            ]
        )

    @classmethod
    def from_bytes(cls, msg_bytes):
        return cls(msg_bytes[1], msg_bytes[2], channel=msg_bytes[0] & cls.CHANNELMASK)


NoteOff.register_message_type()



# SPDX-FileCopyrightText: 2021 ladyada for Adafruit Industries
# SPDX-License-Identifier: MIT
# simple_test
import time
import random
import usb_midi
import adafruit_midi
from adafruit_midi.control_change import ControlChange
from adafruit_midi.note_off import NoteOff
from adafruit_midi.note_on import NoteOn
from adafruit_midi.pitch_bend import PitchBend

print(usb_midi.ports)
midi = adafruit_midi.MIDI(
    midi_in=usb_midi.ports[0], in_channel=0, midi_out=usb_midi.ports[1], out_channel=0
)
print("Midi test")
# Convert channel numbers at the presentation layer to the ones musicians use
print("Default output channel:", midi.out_channel + 1)
print("Listening on input channel:", midi.in_channel + 1)
while True:
    midi.send(NoteOn(44, 120))  # G sharp 2nd octave
    time.sleep(0.25)
    a_pitch_bend = PitchBend(random.randint(0, 16383))
    midi.send(a_pitch_bend)
    # note how a list of messages can be used
    midi.send([NoteOff("G#2", 120), ControlChange(3, 44)])
    time.sleep(0.5)
    msg = midi.receive()
    if msg is not None:
        print("Received:", msg, "at", time.monotonic())



# SPDX-FileCopyrightText: 2021 Raphaël Doursenaud
#
# SPDX-License-Identifier: MIT

"""
`adafruit_midi.mtc_quarter_frame`
================================================================================

MIDI Time Code (MTC) Quarter Frame message.


* Author(s): Raphaël Doursenaud

Implementation Notes
--------------------

Based upon the official MMA0001 / RP004 / RP008 v4.2.1 MIDI Time Code Specification

"""

from adafruit_midi.midi_message import MIDIMessage

__version__ = "0.0.0-auto.0"
__repo__ = "https://github.com/adafruit/Adafruit_CircuitPython_MIDI.git"


class MtcQuarterFrame(MIDIMessage):
    """MIDI Time Code (MTC) Quarter Frame message.

    :param msgtype: The quarter frame message type:

        0. Frame count LS nibble
        1. Frame count MS nibble
        2. Seconds count LS nibble
        3. Seconds count MS nibble
        4. Minutes count LS nibble
        5. Minutes count MS nibble
        6. Hours count LS nibble
        7. Hours count MS nibble and SMPTE Type

    :param value: The quarter frame value for the specified type.
    """

    _STATUS = 0xF1
    _STATUSMASK = 0xFF
    LENGTH = 2

    def __init__(self, msgtype, value):
        self.type = msgtype
        self.value = value
        super().__init__()
        if not 0 <= self.type <= 7 or not 0 <= self.value <= 0x0F:
            self._raise_valueerror_oor()

    def __bytes__(self):
        return bytes(
            [
                self._STATUS,
                (self.type << 4) + self.value,  # Assemble low and high nibbles
            ]
        )

    @classmethod
    def from_bytes(cls, msg_bytes):
        return cls(msg_bytes[1] >> 4, msg_bytes[1] & 15)  # High nibble  # Low nibble


MtcQuarterFrame.register_message_type()



# SPDX-FileCopyrightText: 2021 ladyada for Adafruit Industries
# SPDX-License-Identifier: MIT

import time
import usb_midi
import adafruit_midi

# A subset of messages/events
# pylint: disable=unused-import
from adafruit_midi.timing_clock import TimingClock

# from adafruit_midi.channel_pressure        import ChannelPressure
from adafruit_midi.control_change import ControlChange
from adafruit_midi.note_off import NoteOff
from adafruit_midi.note_on import NoteOn
from adafruit_midi.pitch_bend import PitchBend


# 0 is MIDI channel 1
midi = adafruit_midi.MIDI(midi_in=usb_midi.ports[0], in_channel=0)

print("Midi input test with pauses")

# Convert channel numbers at the presentation layer to the ones musicians use
print("Input channel:", midi.in_channel + 1)

# play with the pause to simulate code doing other stuff
# in the loop
pauses = [0] * 10 + [0.010] * 10 + [0.100] * 10 + [1.0] * 10

while True:
    for pause in pauses:
        msg = midi.receive()
        if msg is not None:
            print(time.monotonic(), msg)
        if pause:
            time.sleep(pause)



# SPDX-FileCopyrightText: 2019 Kevin J. Walters for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""
`adafruit_midi.midi_message`
================================================================================

An abstract class for objects which represent MIDI messages (events).
When individual messages are imported they register themselves with
:func:register_message_type which makes them recognised
by the parser, :func:from_message_bytes.

Large messages like :class:SystemExclusive can only be parsed if they fit
within the input buffer in :class:MIDI.


* Author(s): Kevin J. Walters

Implementation Notes
--------------------

"""

__version__ = "0.0.0-auto.0"
__repo__ = "https://github.com/adafruit/Adafruit_CircuitPython_MIDI.git"

# From C3 - A and B are above G
# Semitones     A   B   C   D   E   F   G
NOTE_OFFSET = [21, 23, 12, 14, 16, 17, 19]


def channel_filter(channel, channel_spec):
    """
    Utility function to return True iff the given channel matches channel_spec.
    """
    if isinstance(channel_spec, int):
        return channel == channel_spec
    if isinstance(channel_spec, tuple):
        return channel in channel_spec
    raise ValueError("Incorrect type for channel_spec" + str(type(channel_spec)))


def note_parser(note):
    """If note is a string then it will be parsed and converted to a MIDI note (key) number, e.g.
    "C4" will return 60, "C#4" will return 61. If note is not a string it will simply be returned.

    :param note: Either 0-127 int or a str representing the note, e.g. "C#4"
    """
    midi_note = note
    if isinstance(note, str):
        if len(note) < 2:
            raise ValueError("Bad note format")
        noteidx = ord(note[0].upper()) - 65  # 65 os ord('A')
        if not 0 <= noteidx <= 6:
            raise ValueError("Bad note")
        sharpen = 0
        if note[1] == "#":
            sharpen = 1
        elif note[1] == "b":
            sharpen = -1
        # int may throw exception here
        midi_note = int(note[1 + abs(sharpen) :]) * 12 + NOTE_OFFSET[noteidx] + sharpen

    return midi_note


class MIDIMessage:
    """
    The parent class for MIDI messages.

    Class variables:

      * ``_STATUS`` - extracted from status byte with channel replaced by 0s
        (high bit is always set by convention).
      * ``_STATUSMASK`` - mask used to compared a status byte with ``_STATUS`` value.
      * ``LENGTH`` - length for a fixed size message *including* status
        or -1 for variable length.
      * ``CHANNELMASK`` - mask used to apply a (wire protocol) channel number.
      * ``ENDSTATUS`` - the end of message status byte, only set for variable length.

    This is an *abstract* class.
    """

    _STATUS = None
    _STATUSMASK = None
    LENGTH = None
    CHANNELMASK = 0x0F
    ENDSTATUS = None

    # Commonly used exceptions to save memory
    @staticmethod
    def _raise_valueerror_oor():
        raise ValueError("Out of range")

    # Each element is ((status, mask), class)
    # order is more specific masks first
    _statusandmask_to_class = []

    def __init__(self, *, channel=None):
        self._channel = channel  # dealing with pylint inadequacy
        self.channel = channel

    @property
    def channel(self):
        """The channel number of the MIDI message where appropriate.
        This is *updated* by MIDI.send() method.
        """
        return self._channel

    @channel.setter
    def channel(self, channel):
        if channel is not None and not 0 <= channel <= 15:
            raise ValueError("Channel must be 0-15 or None")
        self._channel = channel

    @classmethod
    def register_message_type(cls):
        """Register a new message by its status value and mask.
        This is called automagically at ``import`` time for each message.
        """
        ### These must be inserted with more specific masks first
        insert_idx = len(MIDIMessage._statusandmask_to_class)
        for idx, m_type in enumerate(MIDIMessage._statusandmask_to_class):
            if cls._STATUSMASK > m_type[0][1]:
                insert_idx = idx
                break

        MIDIMessage._statusandmask_to_class.insert(
            insert_idx, ((cls._STATUS, cls._STATUSMASK), cls)
        )

    # pylint: disable=too-many-arguments
    @classmethod
    def _search_eom_status(cls, buf, eom_status, msgstartidx, msgendidxplusone, endidx):
        good_termination = False
        bad_termination = False

        msgendidxplusone = msgstartidx + 1
        while msgendidxplusone <= endidx:
            # Look for a status byte
            # Second rule of the MIDI club is status bytes have MSB set
            if buf[msgendidxplusone] & 0x80:
                # pylint: disable=simplifiable-if-statement
                if buf[msgendidxplusone] == eom_status:
                    good_termination = True
                else:
                    bad_termination = True
                break
            msgendidxplusone += 1

        if good_termination or bad_termination:
            msgendidxplusone += 1

        return (msgendidxplusone, good_termination, bad_termination)

    @classmethod
    def _match_message_status(cls, buf, msgstartidx, msgendidxplusone, endidx):
        msgclass = None
        status = buf[msgstartidx]
        known_msg = False
        complete_msg = False
        bad_termination = False

        # Rummage through our list looking for a status match
        for status_mask, msgclass in MIDIMessage._statusandmask_to_class:
            masked_status = status & status_mask[1]
            if status_mask[0] == masked_status:
                known_msg = True
                # Check there's enough left to parse a complete message
                # this value can be changed later for a var. length msgs
                complete_msg = len(buf) - msgstartidx >= msgclass.LENGTH
                if not complete_msg:
                    break

                if msgclass.LENGTH < 0:  # indicator of variable length message
                    (
                        msgendidxplusone,
                        terminated_msg,
                        bad_termination,
                    ) = cls._search_eom_status(
                        buf, msgclass.ENDSTATUS, msgstartidx, msgendidxplusone, endidx
                    )
                    if not terminated_msg:
                        complete_msg = False
                else:  # fixed length message
                    msgendidxplusone = msgstartidx + msgclass.LENGTH
                break

        return (
            msgclass,
            status,
            known_msg,
            complete_msg,
            bad_termination,
            msgendidxplusone,
        )

    # pylint: disable=too-many-locals,too-many-branches
    @classmethod
    def from_message_bytes(cls, midibytes, channel_in):
        """Create an appropriate object of the correct class for the
        first message found in some MIDI bytes filtered by channel_in.

        Returns (messageobject, endplusone, skipped)
        or for no messages, partial messages or messages for other channels
        (None, endplusone, skipped).
        """
        endidx = len(midibytes) - 1
        skipped = 0
        preamble = True

        msgstartidx = 0
        msgendidxplusone = 0
        while True:
            msg = None
            # Look for a status byte
            # Second rule of the MIDI club is status bytes have MSB set
            while msgstartidx <= endidx and not midibytes[msgstartidx] & 0x80:
                msgstartidx += 1
                if preamble:
                    skipped += 1
            preamble = False

            # Either no message or a partial one
            if msgstartidx > endidx:
                return (None, endidx + 1, skipped)

            # Try and match the status byte found in midibytes
            (
                msgclass,
                status,
                known_message,
                complete_message,
                bad_termination,
                msgendidxplusone,
            ) = cls._match_message_status(
                midibytes, msgstartidx, msgendidxplusone, endidx
            )
            channel_match_orna = True
            if complete_message and not bad_termination:
                try:
                    msg = msgclass.from_bytes(midibytes[msgstartidx:msgendidxplusone])
                    if msg.channel is not None:
                        channel_match_orna = channel_filter(msg.channel, channel_in)

                except (ValueError, TypeError) as ex:
                    msg = MIDIBadEvent(midibytes[msgstartidx:msgendidxplusone], ex)

            # break out of while loop for a complete message on good channel
            # or we have one we do not know about
            if known_message:
                if complete_message:
                    if channel_match_orna:
                        break
                    # advance to next message
                    msgstartidx = msgendidxplusone
                else:
                    # Important case of a known message but one that is not
                    # yet complete - leave bytes in buffer and wait for more
                    break
            else:
                msg = MIDIUnknownEvent(status)
                # length cannot be known
                # next read will skip past leftover data bytes
                msgendidxplusone = msgstartidx + 1
                break

        return (msg, msgendidxplusone, skipped)

    # A default method for constructing wire messages with no data.
    # Returns an (immutable) bytes with just the status code in.
    def __bytes__(self):
        """Return the ``bytes`` wire protocol representation of the object
        with channel number applied where appropriate."""
        return bytes([self._STATUS])

    # databytes value present to keep interface uniform but unused
    # A default method for constructing message objects with no data.
    # Returns the new object.
    # pylint: disable=unused-argument
    @classmethod
    def from_bytes(cls, msg_bytes):
        """Creates an object from the byte stream of the wire protocol
        representation of the MIDI message."""
        return cls()


# DO NOT try to register these messages
class MIDIUnknownEvent(MIDIMessage):
    """An unknown MIDI message.

    :param int status: The MIDI status number.

    This can either occur because there is no class representing the message
    or because it is not imported.
    """

    LENGTH = -1

    def __init__(self, status):
        self.status = status
        super().__init__()


class MIDIBadEvent(MIDIMessage):
    """A bad MIDI message, one that could not be parsed/constructed.

    :param list msg_bytes: The MIDI status including any embedded channel number
        and associated subsequent data bytes.
    :param Exception exception: The exception used to store the repr() text representation.

    This could be due to status bytes appearing where data bytes are expected.
    The channel property will not be set.
    """

    LENGTH = -1

    def __init__(self, msg_bytes, exception):
        self.data = bytes(msg_bytes)
        self.exception_text = repr(exception)
        super().__init__()



# SPDX-FileCopyrightText: 2019 Kevin J. Walters for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""
`adafruit_midi.note_on`
================================================================================

Note On Change MIDI message.


* Author(s): Kevin J. Walters

Implementation Notes
--------------------

"""

from .midi_message import MIDIMessage, note_parser

__version__ = "0.0.0-auto.0"
__repo__ = "https://github.com/adafruit/Adafruit_CircuitPython_MIDI.git"


class NoteOn(MIDIMessage):
    """Note On Change MIDI message.

    :param note: The note (key) number either as an ``int`` (0-127) or a
        ``str`` which is parsed, e.g. "C4" (middle C) is 60, "A4" is 69.
    :param int velocity: The strike velocity, 0-127, 0 is equivalent
        to a Note Off, defaults to 127.
    """

    _STATUS = 0x90
    _STATUSMASK = 0xF0
    LENGTH = 3

    def __init__(self, note, velocity=127, *, channel=None):
        self.note = note_parser(note)
        """Key, either int (0-127) or string that will be turned on """
        self.velocity = velocity
        """Strike velocity, int (0-127); 0 is equivalent to Note Off """
        super().__init__(channel=channel)
        if not 0 <= self.note <= 127 or not 0 <= self.velocity <= 127:
            self._raise_valueerror_oor()

    def __bytes__(self):
        return bytes(
            [self._STATUS | (self.channel & self.CHANNELMASK), self.note, self.velocity]
        )

    @classmethod
    def from_bytes(cls, msg_bytes):
        return cls(msg_bytes[1], msg_bytes[2], channel=msg_bytes[0] & cls.CHANNELMASK)


NoteOn.register_message_type()



# pylint: disable=invalid-name
# SPDX-FileCopyrightText: 2019 Kevin J. Walters for Adafruit Industries
#
# SPDX-License-Identifier: MIT
# pylint: enable=invalid-name

import unittest
from unittest.mock import Mock, call

import random
import os

verbose = int(os.getenv("TESTVERBOSE", "2"))

# pylint: disable=wrong-import-position
# adafruit_midi had an import usb_midi
import sys

# sys.modules['usb_midi'] = MagicMock()

# Borrowing the dhalbert/tannewt technique from adafruit/Adafruit_CircuitPython_Motor
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

# Full monty
from adafruit_midi.channel_pressure import ChannelPressure
from adafruit_midi.control_change import ControlChange
from adafruit_midi.note_off import NoteOff
from adafruit_midi.note_on import NoteOn
from adafruit_midi.pitch_bend import PitchBend
from adafruit_midi.system_exclusive import SystemExclusive

# Import after messages - opposite to other test file
import adafruit_midi

# pylint: enable=wrong-import-position


# For loopback/echo tests
def MIDI_mocked_both_loopback(in_c, out_c):  # pylint: disable=invalid-name
    # pylint: enable=invalid-name
    usb_data = bytearray()

    def write(buffer, length):
        nonlocal usb_data
        usb_data.extend(buffer[0:length])

    def read(length):
        nonlocal usb_data
        poppedbytes = usb_data[0:length]
        usb_data = usb_data[len(poppedbytes) :]
        return bytes(poppedbytes)

    mockedportin = Mock()
    mockedportin.read = read
    mockedportout = Mock()
    mockedportout.write = write
    midi = adafruit_midi.MIDI(
        midi_out=mockedportout, midi_in=mockedportin, out_channel=out_c, in_channel=in_c
    )
    return midi


def MIDI_mocked_receive(in_c, data, read_sizes):  # pylint: disable=invalid-name
    # pylint: enable=invalid-name
    usb_data = bytearray(data)
    chunks = read_sizes
    chunk_idx = 0

    def read(length):
        nonlocal usb_data, chunks, chunk_idx
        # pylint:  disable=no-else-return
        if length != 0 and chunk_idx < len(chunks):
            # min() to ensure we only read what's asked for and present
            poppedbytes = usb_data[0 : min(length, chunks[chunk_idx])]
            usb_data = usb_data[len(poppedbytes) :]
            if length >= chunks[chunk_idx]:
                chunk_idx += 1
            else:
                chunks[chunk_idx] -= len(poppedbytes)
            return bytes(poppedbytes)
        else:
            return bytes()

    mockedportin = Mock()
    mockedportin.read = read

    midi = adafruit_midi.MIDI(
        midi_out=None, midi_in=mockedportin, out_channel=in_c, in_channel=in_c
    )
    return midi


class Test_MIDI_constructor(unittest.TestCase):  # pylint: disable=invalid-name
    # pylint: enable=invalid-name
    def test_no_inout(self):
        # constructor likes a bit of in out
        with self.assertRaises(ValueError):
            adafruit_midi.MIDI()


class Test_MIDI(unittest.TestCase):
    # pylint: disable=too-many-branches
    def test_captured_data_one_byte_reads(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        channel = 0
        # From an M-Audio AXIOM controller
        raw_data = bytearray(
            [0x90, 0x3E, 0x5F]
            + [0xD0, 0x10]
            + [0x90, 0x40, 0x66]
            + [0xB0, 0x1, 0x08]
            + [0x90, 0x41, 0x74]
            + [0xE0, 0x03, 0x40]
        )
        midi = MIDI_mocked_receive(channel, raw_data, [1] * len(raw_data))

        for unused in range(100):  # pylint: disable=unused-variable
            msg = midi.receive()
            if msg is not None:
                break
        self.assertIsInstance(msg, NoteOn)
        self.assertEqual(msg.note, 0x3E)
        self.assertEqual(msg.velocity, 0x5F)
        self.assertEqual(msg.channel, channel)

        # for loops currently absorb any Nones but could
        # be set to read precisely the expected number...
        for unused in range(100):  # pylint: disable=unused-variable
            msg = midi.receive()
            if msg is not None:
                break
        self.assertIsInstance(msg, ChannelPressure)
        self.assertEqual(msg.pressure, 0x10)
        self.assertEqual(msg.channel, channel)

        for unused in range(100):  # pylint: disable=unused-variable
            msg = midi.receive()
            if msg is not None:
                break
        self.assertIsInstance(msg, NoteOn)
        self.assertEqual(msg.note, 0x40)
        self.assertEqual(msg.velocity, 0x66)
        self.assertEqual(msg.channel, channel)

        for unused in range(100):  # pylint: disable=unused-variable
            msg = midi.receive()
            if msg is not None:
                break
        self.assertIsInstance(msg, ControlChange)
        self.assertEqual(msg.control, 0x01)
        self.assertEqual(msg.value, 0x08)
        self.assertEqual(msg.channel, channel)

        for unused in range(100):  # pylint: disable=unused-variable
            msg = midi.receive()
            if msg is not None:
                break
        self.assertIsInstance(msg, NoteOn)
        self.assertEqual(msg.note, 0x41)
        self.assertEqual(msg.velocity, 0x74)
        self.assertEqual(msg.channel, channel)

        for unused in range(100):  # pylint: disable=unused-variable
            msg = midi.receive()
            if msg is not None:
                break
        self.assertIsInstance(msg, PitchBend)
        self.assertEqual(msg.pitch_bend, 8195)
        self.assertEqual(msg.channel, channel)

        for unused in range(100):  # pylint: disable=unused-variable
            msg = midi.receive()
            self.assertIsNone(msg)

    def test_unknown_before_NoteOn(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        channel = 0
        # From an M-Audio AXIOM controller
        raw_data = bytes(
            [0b11110011, 0x10]  # Song Select (not yet implemented)
            + [0b11110011, 0x20]
            + [0b11110100]
            + [0b11110101]
        ) + bytes(NoteOn("C5", 0x7F, channel=channel))
        midi = MIDI_mocked_receive(channel, raw_data, [2, 2, 1, 1, 3])

        for unused in range(4):  # pylint: disable=unused-variable
            msg = midi.receive()
            self.assertIsInstance(msg, adafruit_midi.midi_message.MIDIUnknownEvent)
            self.assertIsNone(msg.channel)

        msg = midi.receive()
        self.assertIsInstance(msg, NoteOn)
        self.assertEqual(msg.note, 0x48)  # 0x48 is C5
        self.assertEqual(msg.velocity, 0x7F)
        self.assertEqual(msg.channel, channel)

    # See https://github.com/adafruit/Adafruit_CircuitPython_MIDI/issues/8
    def test_running_status_when_implemented(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        channel = 8
        raw_data = (
            bytes(NoteOn("C5", 0x7F, channel=channel))
            + bytes([0xE8, 0x72, 0x40] + [0x6D, 0x40] + [0x05, 0x41])
            + bytes(NoteOn("D5", 0x7F, channel=channel))
        )

        midi = MIDI_mocked_receive(channel, raw_data, [3 + 3 + 2 + 3 + 3])
        self.assertIsInstance(midi, adafruit_midi.MIDI)  # silence pylint!
        # self.assertEqual(TOFINISH, WHENIMPLEMENTED)

    def test_somegood_somemissing_databytes(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        channel = 8
        raw_data = (
            bytes(NoteOn("C5", 0x7F, channel=channel))
            + bytes(
                [0xE8, 0x72, 0x40]
                + [0xE8, 0x6D]  # Missing last data byte
                + [0xE8, 0x5, 0x41]
            )
            + bytes(NoteOn("D5", 0x7F, channel=channel))
        )
        midi = MIDI_mocked_receive(channel, raw_data, [3 + 3 + 2 + 3 + 3])

        msg1 = midi.receive()
        self.assertIsInstance(msg1, NoteOn)
        self.assertEqual(msg1.note, 72)
        self.assertEqual(msg1.velocity, 0x7F)
        self.assertEqual(msg1.channel, channel)

        msg2 = midi.receive()
        self.assertIsInstance(msg2, PitchBend)
        self.assertEqual(msg2.pitch_bend, 8306)
        self.assertEqual(msg2.channel, channel)

        # The current implementation will read status bytes for data
        # In most cases it would be a faster recovery with fewer messages
        # lost if the next status byte wasn't consumed
        # and parsing restarted from that byte
        msg3 = midi.receive()
        self.assertIsInstance(msg3, adafruit_midi.midi_message.MIDIBadEvent)
        self.assertIsInstance(msg3.data, bytes)
        self.assertEqual(msg3.data, bytes([0xE8, 0x6D, 0xE8]))
        self.assertIsNone(msg3.channel)

        # (msg4, channel4) = m.receive()
        # self.assertIsInstance(msg4, PitchBend)
        # self.assertEqual(msg4.pitch_bend, 72)
        # self.assertEqual(channel4, c)

        msg5 = midi.receive()
        self.assertIsInstance(msg5, NoteOn)
        self.assertEqual(msg5.note, 74)
        self.assertEqual(msg5.velocity, 0x7F)
        self.assertEqual(msg5.channel, channel)

        msg6 = midi.receive()
        self.assertIsNone(msg6)

    def test_smallsysex_between_notes(self):
        midi = MIDI_mocked_both_loopback(3, 3)

        midi.send(
            [
                NoteOn("C4", 0x7F),
                SystemExclusive([0x1F], [1, 2, 3, 4, 5, 6, 7, 8]),
                NoteOff(60, 0x28),
            ]
        )

        msg1 = midi.receive()
        self.assertIsInstance(msg1, NoteOn)
        self.assertEqual(msg1.note, 60)
        self.assertEqual(msg1.velocity, 0x7F)
        self.assertEqual(msg1.channel, 3)

        msg2 = midi.receive()
        self.assertIsInstance(msg2, SystemExclusive)
        self.assertEqual(msg2.manufacturer_id, bytes([0x1F]))
        self.assertEqual(msg2.data, bytes([1, 2, 3, 4, 5, 6, 7, 8]))
        self.assertEqual(msg2.channel, None)  # SysEx does not have a channel

        msg3 = midi.receive()
        self.assertIsInstance(msg3, NoteOff)
        self.assertEqual(msg3.note, 60)
        self.assertEqual(msg3.velocity, 0x28)
        self.assertEqual(msg3.channel, 3)

        msg4 = midi.receive()
        self.assertIsNone(msg4)

    def test_smallsysex_bytes_type(self):
        message = SystemExclusive([0x1F], [100, 150, 200])

        self.assertIsInstance(message, SystemExclusive)
        self.assertEqual(message.manufacturer_id, bytes([0x1F]))
        self.assertIsInstance(message.manufacturer_id, bytes)

        # check this really is immutable (pylint also picks this up!)
        with self.assertRaises(TypeError):
            message.data[0] = 0  # pylint: disable=unsupported-assignment-operation

        self.assertEqual(message.data, bytes([100, 150, 200]))
        self.assertIsInstance(message.data, bytes)

    # pylint: disable=too-many-locals
    def test_larger_than_buffer_sysex(self):
        channel = 0
        monster_data_len = 500
        raw_data = (
            bytes(NoteOn("C5", 0x7F, channel=channel))
            + bytes(
                SystemExclusive([0x02], [d & 0x7F for d in range(monster_data_len)])
            )
            + bytes(NoteOn("D5", 0x7F, channel=channel))
        )
        midi = MIDI_mocked_receive(channel, raw_data, [len(raw_data)])
        buffer_len = midi._in_buf_size  # pylint: disable=protected-access
        self.assertTrue(
            monster_data_len > buffer_len,
            "checking our SysEx truly is larger than buffer",
        )

        msg1 = midi.receive()
        self.assertIsInstance(msg1, NoteOn)
        self.assertEqual(msg1.note, 72)
        self.assertEqual(msg1.velocity, 0x7F)
        self.assertEqual(msg1.channel, channel)

        # (Ab)using python's rounding down for negative division
        # pylint: disable=unused-variable
        for unused in range(-(-(1 + 1 + monster_data_len + 1) // buffer_len) - 1):
            msg2 = midi.receive()
            self.assertIsNone(msg2)

        # The current implementation will read SysEx end status byte
        # and report it as an unknown
        msg3 = midi.receive()
        self.assertIsInstance(msg3, adafruit_midi.midi_message.MIDIUnknownEvent)
        self.assertEqual(msg3.status, 0xF7)
        self.assertIsNone(msg3.channel)

        # (msg4, channel4) = m.receive()
        # self.assertIsInstance(msg4, PitchBend)
        # self.assertEqual(msg4.pitch_bend, 72)
        # self.assertEqual(channel4, c)

        msg5 = midi.receive()
        self.assertIsInstance(msg5, NoteOn)
        self.assertEqual(msg5.note, 74)
        self.assertEqual(msg5.velocity, 0x7F)
        self.assertEqual(msg5.channel, channel)

        msg6 = midi.receive()
        self.assertIsNone(msg6)


# pylint does not like mock_calls - must be a better way to handle this?
# pylint: disable=no-member
class Test_MIDI_send(unittest.TestCase):
    def test_send_basic_single(self):
        # def printit(buffer, len):
        #    print(buffer[0:len])
        mockedportout = Mock()
        # mockedPortOut.write = printit

        midi = adafruit_midi.MIDI(midi_out=mockedportout, out_channel=2)

        # Test sending some NoteOn and NoteOff to various channels
        nextcall = 0
        midi.send(NoteOn(0x60, 0x7F))
        self.assertEqual(
            mockedportout.write.mock_calls[nextcall], call(b"\x92\x60\x7f", 3)
        )
        nextcall += 1
        midi.send(NoteOn(0x64, 0x3F))
        self.assertEqual(
            mockedportout.write.mock_calls[nextcall], call(b"\x92\x64\x3f", 3)
        )
        nextcall += 1
        midi.send(NoteOn(0x67, 0x1F))
        self.assertEqual(
            mockedportout.write.mock_calls[nextcall], call(b"\x92\x67\x1f", 3)
        )
        nextcall += 1

        midi.send(NoteOn(0x60, 0x00))  # Alternative to NoteOff
        self.assertEqual(
            mockedportout.write.mock_calls[nextcall], call(b"\x92\x60\x00", 3)
        )
        nextcall += 1
        midi.send(NoteOff(0x64, 0x01))
        self.assertEqual(
            mockedportout.write.mock_calls[nextcall], call(b"\x82\x64\x01", 3)
        )
        nextcall += 1
        midi.send(NoteOff(0x67, 0x02))
        self.assertEqual(
            mockedportout.write.mock_calls[nextcall], call(b"\x82\x67\x02", 3)
        )
        nextcall += 1

        # Setting channel to non default
        midi.send(NoteOn(0x6C, 0x7F), channel=9)
        self.assertEqual(
            mockedportout.write.mock_calls[nextcall], call(b"\x99\x6c\x7f", 3)
        )
        nextcall += 1

        midi.send(NoteOff(0x6C, 0x7F), channel=9)
        self.assertEqual(
            mockedportout.write.mock_calls[nextcall], call(b"\x89\x6c\x7f", 3)
        )
        nextcall += 1

    def test_send_badnotes(self):
        mockedportout = Mock()

        midi = adafruit_midi.MIDI(midi_out=mockedportout, out_channel=2)

        # Test sending some NoteOn and NoteOff to various channels
        nextcall = 0
        midi.send(NoteOn(60, 0x7F))
        self.assertEqual(
            mockedportout.write.mock_calls[nextcall], call(b"\x92\x3c\x7f", 3)
        )
        nextcall += 1
        with self.assertRaises(ValueError):
            midi.send(NoteOn(64, 0x80))  # Velocity > 127 - illegal value

        with self.assertRaises(ValueError):
            midi.send(NoteOn(67, -1))

        # test after exceptions to ensure sending is still ok
        midi.send(NoteOn(72, 0x7F))
        self.assertEqual(
            mockedportout.write.mock_calls[nextcall], call(b"\x92\x48\x7f", 3)
        )
        nextcall += 1

    def test_send_basic_sequences(self):
        # def printit(buffer, len):
        #    print(buffer[0:len])
        mockedportout = Mock()
        # mockedportout.write = printit

        midi = adafruit_midi.MIDI(midi_out=mockedportout, out_channel=2)

        # Test sending some NoteOn and NoteOff to various channels
        nextcall = 0
        # Test sequences with list syntax and pass a tuple too
        note_list = [NoteOn(0x6C, 0x51), NoteOn(0x70, 0x52), NoteOn(0x73, 0x53)]
        note_tuple = tuple(note_list)
        midi.send(note_list, channel=10)
        self.assertEqual(
            mockedportout.write.mock_calls[nextcall],
            call(b"\x9a\x6c\x51\x9a\x70\x52\x9a\x73\x53", 9),
            "The implementation writes in one go, single 9 byte write expected",
        )
        nextcall += 1
        midi.send(note_tuple, channel=11)
        self.assertEqual(
            mockedportout.write.mock_calls[nextcall],
            call(b"\x9b\x6c\x51\x9b\x70\x52\x9b\x73\x53", 9),
            "The implementation writes in one go, single 9 byte write expected",
        )
        nextcall += 1

    def test_termination_with_random_data(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        """Test with a random stream of bytes to ensure that the parsing code
        termates and returns, i.e. does not go into any infinite loops.
        """
        channel = 0
        random.seed(303808)
        raw_data = bytearray([random.randint(0, 255) for i in range(50000)])
        midi = MIDI_mocked_receive(channel, raw_data, [len(raw_data)])

        noinfiniteloops = False
        for unused in range(len(raw_data)):  # pylint: disable=unused-variable
            midi.receive()  # not interested in returned tuple

        noinfiniteloops = True  # interested in getting to here
        self.assertTrue(noinfiniteloops)


if __name__ == "__main__":
    unittest.main(verbosity=verbose)



# SPDX-FileCopyrightText: 2019 Kevin J. Walters for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""
`adafruit_midi.program_change`
================================================================================

Program Change MIDI message.


* Author(s): Kevin J. Walters

Implementation Notes
--------------------

"""

from .midi_message import MIDIMessage

__version__ = "0.0.0-auto.0"
__repo__ = "https://github.com/adafruit/Adafruit_CircuitPython_MIDI.git"


class ProgramChange(MIDIMessage):
    """Program Change MIDI message.

    :param int patch: The new program/patch number to use, 0-127.
    """

    _STATUS = 0xC0
    _STATUSMASK = 0xF0
    LENGTH = 2

    def __init__(self, patch, *, channel=None):
        self.patch = patch
        super().__init__(channel=channel)
        if not 0 <= self.patch <= 127:
            self._raise_valueerror_oor()

    def __bytes__(self):
        return bytes([self._STATUS | (self.channel & self.CHANNELMASK), self.patch])

    @classmethod
    def from_bytes(cls, msg_bytes):
        return cls(msg_bytes[1], channel=msg_bytes[0] & cls.CHANNELMASK)


ProgramChange.register_message_type()



# pylint: disable=invalid-name
# SPDX-FileCopyrightText: 2019 Kevin J. Walters for Adafruit Industries
#
# SPDX-License-Identifier: MIT
# pylint: enable=invalid-name

import unittest


import os

verbose = int(os.getenv("TESTVERBOSE", "2"))

# pylint: disable=wrong-import-position
# adafruit_midi had an import usb_midi
import sys

# sys.modules['usb_midi'] = MagicMock()

# Borrowing the dhalbert/tannewt technique from adafruit/Adafruit_CircuitPython_Motor
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

# Import before messages - opposite to other test file
import adafruit_midi

# Full monty
from adafruit_midi.note_off import NoteOff
from adafruit_midi.note_on import NoteOn
from adafruit_midi.system_exclusive import SystemExclusive

# pylint: enable=wrong-import-position

# pylint: disable=invalid-name
class Test_MIDIMessage_from_message_byte_tests(unittest.TestCase):
    # pylint: enable=invalid-name
    def test_NoteOn_basic(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        data = bytes([0x90, 0x30, 0x7F])
        ichannel = 0

        (msg, msgendidxplusone, skipped) = adafruit_midi.MIDIMessage.from_message_bytes(
            data, ichannel
        )

        self.assertIsInstance(msg, NoteOn)
        self.assertEqual(msg.note, 0x30)
        self.assertEqual(msg.velocity, 0x7F)
        self.assertEqual(msgendidxplusone, 3)
        self.assertEqual(skipped, 0)
        self.assertEqual(msg.channel, 0)

    def test_NoteOn_awaitingthirdbyte(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        data = bytes([0x90, 0x30])
        ichannel = 0

        (msg, msgendidxplusone, skipped) = adafruit_midi.MIDIMessage.from_message_bytes(
            data, ichannel
        )
        self.assertIsNone(msg)
        self.assertEqual(
            msgendidxplusone,
            skipped,
            "skipped must be 0 as it only indicates bytes before a status byte",
        )
        self.assertEqual(
            msgendidxplusone,
            0,
            "msgendidxplusone must be 0 as buffer must be lest as is for more data",
        )
        self.assertEqual(skipped, 0)

    def test_NoteOn_predatajunk(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        data = bytes([0x20, 0x64, 0x90, 0x30, 0x32])
        ichannel = 0

        (msg, msgendidxplusone, skipped) = adafruit_midi.MIDIMessage.from_message_bytes(
            data, ichannel
        )

        self.assertIsInstance(msg, NoteOn)
        self.assertEqual(msg.note, 0x30)
        self.assertEqual(msg.velocity, 0x32)
        self.assertEqual(
            msgendidxplusone,
            5,
            "data bytes from partial message and messages are removed",
        )
        self.assertEqual(skipped, 2)
        self.assertEqual(msg.channel, 0)

    def test_NoteOn_prepartialsysex(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        data = bytes([0x01, 0x02, 0x03, 0x04, 0xF7, 0x90, 0x30, 0x32])
        ichannel = 0

        (msg, msgendidxplusone, skipped) = adafruit_midi.MIDIMessage.from_message_bytes(
            data, ichannel
        )

        # MIDIMessage parsing could be improved to return something that
        # indicates its a truncated end of SysEx
        self.assertIsInstance(msg, adafruit_midi.midi_message.MIDIUnknownEvent)
        self.assertEqual(msg.status, 0xF7)
        self.assertEqual(
            msgendidxplusone,
            5,
            "removal of the end of the partial SysEx data and terminating status byte",
        )
        self.assertEqual(skipped, 4, "skipped only counts data bytes so will be 4 here")
        self.assertIsNone(msg.channel)

        data = data[msgendidxplusone:]
        (msg, msgendidxplusone, skipped) = adafruit_midi.MIDIMessage.from_message_bytes(
            data, ichannel
        )

        self.assertIsInstance(
            msg,
            NoteOn,
            "NoteOn is expected if SystemExclusive is loaded otherwise it'd be MIDIUnknownEvent",
        )
        self.assertEqual(msg.note, 0x30)
        self.assertEqual(msg.velocity, 0x32)
        self.assertEqual(msgendidxplusone, 3, "NoteOn message removed")
        self.assertEqual(skipped, 0)
        self.assertEqual(msg.channel, 0)

    def test_NoteOn_postNoteOn(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        data = bytes([0x90 | 0x08, 0x30, 0x7F, 0x90 | 0x08, 0x37, 0x64])
        ichannel = 8

        (msg, msgendidxplusone, skipped) = adafruit_midi.MIDIMessage.from_message_bytes(
            data, ichannel
        )

        self.assertIsInstance(msg, NoteOn)
        self.assertEqual(msg.note, 0x30)
        self.assertEqual(msg.velocity, 0x7F)
        self.assertEqual(msgendidxplusone, 3)
        self.assertEqual(skipped, 0)
        self.assertEqual(msg.channel, 8)

    def test_NoteOn_postpartialNoteOn(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        data = bytes([0x90, 0x30, 0x7F, 0x90, 0x37])
        ichannel = 0

        (msg, msgendidxplusone, skipped) = adafruit_midi.MIDIMessage.from_message_bytes(
            data, ichannel
        )

        self.assertIsInstance(msg, NoteOn)
        self.assertEqual(msg.note, 0x30)
        self.assertEqual(msg.velocity, 0x7F)
        self.assertEqual(msgendidxplusone, 3, "Only first message is removed")
        self.assertEqual(skipped, 0)
        self.assertEqual(msg.channel, 0)

    def test_NoteOn_preotherchannel(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        data = bytes([0x90 | 0x05, 0x30, 0x7F, 0x90 | 0x03, 0x37, 0x64])
        ichannel = 3

        (msg, msgendidxplusone, skipped) = adafruit_midi.MIDIMessage.from_message_bytes(
            data, ichannel
        )

        self.assertIsInstance(msg, NoteOn)
        self.assertEqual(msg.note, 0x37)
        self.assertEqual(msg.velocity, 0x64)
        self.assertEqual(msgendidxplusone, 6, "Both messages are removed from buffer")
        self.assertEqual(skipped, 0)
        self.assertEqual(msg.channel, 3)

    def test_NoteOn_preotherchannelplusintermediatejunk(
        self,
    ):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        data = bytes([0x90 | 0x05, 0x30, 0x7F, 0x00, 0x00, 0x90 | 0x03, 0x37, 0x64])
        ichannel = 3

        (msg, msgendidxplusone, skipped) = adafruit_midi.MIDIMessage.from_message_bytes(
            data, ichannel
        )

        self.assertIsInstance(msg, NoteOn)
        self.assertEqual(msg.note, 0x37)
        self.assertEqual(msg.velocity, 0x64)
        self.assertEqual(
            msgendidxplusone, 8, "Both messages and junk are removed from buffer"
        )
        self.assertEqual(skipped, 0)
        self.assertEqual(msg.channel, 3)

    def test_NoteOn_wrongchannel(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        data = bytes([0x95, 0x30, 0x7F])
        ichannel = 3

        (msg, msgendidxplusone, skipped) = adafruit_midi.MIDIMessage.from_message_bytes(
            data, ichannel
        )

        self.assertIsNone(msg)
        self.assertEqual(msgendidxplusone, 3, "wrong channel message discarded")
        self.assertEqual(skipped, 0)

    def test_NoteOn_partialandpreotherchannel1(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        data = bytes([0x95, 0x30, 0x7F, 0x93])
        ichannel = 3

        (msg, msgendidxplusone, skipped) = adafruit_midi.MIDIMessage.from_message_bytes(
            data, ichannel
        )

        self.assertIsNone(msg)
        self.assertEqual(
            msgendidxplusone, 3, "first message discarded, second partial left"
        )
        self.assertEqual(skipped, 0)

    def test_NoteOn_partialandpreotherchannel2(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        data = bytes([0x95, 0x30, 0x7F, 0x93, 0x37])
        ichannel = 3

        (msg, msgendidxplusone, skipped) = adafruit_midi.MIDIMessage.from_message_bytes(
            data, ichannel
        )

        self.assertIsNone(msg)
        self.assertEqual(
            msgendidxplusone, 3, "first message discarded, second partial left"
        )
        self.assertEqual(skipped, 0)

    def test_NoteOn_constructor_int(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        object1 = NoteOn(60, 0x7F)

        self.assertEqual(object1.note, 60)
        self.assertEqual(object1.velocity, 0x7F)
        self.assertIsNone(object1.channel)

        object2 = NoteOn(60, 0x00)  # equivalent of NoteOff

        self.assertEqual(object2.note, 60)
        self.assertEqual(object2.velocity, 0x00)
        self.assertIsNone(object2.channel)

        object3 = NoteOn(60, 0x50, channel=7)

        self.assertEqual(object3.note, 60)
        self.assertEqual(object3.velocity, 0x50)
        self.assertEqual(object3.channel, 7)

        object4 = NoteOn(60)  # velocity defaults to 127

        self.assertEqual(object4.note, 60)
        self.assertEqual(object4.velocity, 127)
        self.assertIsNone(object4.channel)

    def test_SystemExclusive_NoteOn(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        data = bytes([0xF0, 0x42, 0x01, 0x02, 0x03, 0x04, 0xF7, 0x90 | 14, 0x30, 0x60])
        ichannel = 14

        (msg, msgendidxplusone, skipped) = adafruit_midi.MIDIMessage.from_message_bytes(
            data, ichannel
        )

        self.assertIsInstance(msg, SystemExclusive)
        self.assertEqual(msg.manufacturer_id, bytes([0x42]))  # Korg
        self.assertEqual(msg.data, bytes([0x01, 0x02, 0x03, 0x04]))
        self.assertEqual(msgendidxplusone, 7)
        self.assertEqual(
            skipped, 0, "If SystemExclusive class is imported then this must be 0"
        )
        self.assertIsNone(msg.channel)

        (msg, msgendidxplusone, skipped) = adafruit_midi.MIDIMessage.from_message_bytes(
            data[msgendidxplusone:], ichannel
        )

        self.assertIsInstance(msg, NoteOn)
        self.assertEqual(msg.note, 48)
        self.assertEqual(msg.velocity, 0x60)
        self.assertEqual(msgendidxplusone, 3)
        self.assertEqual(skipped, 0)
        self.assertEqual(msg.channel, 14)

    def test_SystemExclusive_NoteOn_premalterminatedsysex(
        self,
    ):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        data = bytes([0xF0, 0x42, 0x01, 0x02, 0x03, 0x04, 0xF0, 0x90, 0x30, 0x32])
        ichannel = 0

        # 0xf0 is incorrect status to mark end of this message, must be 0xf7
        (msg, msgendidxplusone, skipped) = adafruit_midi.MIDIMessage.from_message_bytes(
            data, ichannel
        )

        self.assertIsNone(msg)
        self.assertEqual(msgendidxplusone, 7)
        self.assertEqual(
            skipped, 0, "If SystemExclusive class is imported then this must be 0"
        )

    def test_Unknown_SinglebyteStatus(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        data = bytes([0xFD])
        ichannel = 0

        (msg, msgendidxplusone, skipped) = adafruit_midi.MIDIMessage.from_message_bytes(
            data, ichannel
        )

        self.assertIsInstance(msg, adafruit_midi.midi_message.MIDIUnknownEvent)
        self.assertEqual(msgendidxplusone, 1)
        self.assertEqual(skipped, 0)
        self.assertIsNone(msg.channel)

    def test_Empty(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        data = bytes([])
        ichannel = 0

        (msg, msgendidxplusone, skipped) = adafruit_midi.MIDIMessage.from_message_bytes(
            data, ichannel
        )

        self.assertIsNone(msg)
        self.assertEqual(msgendidxplusone, 0)
        self.assertEqual(skipped, 0)


class Test_MIDIMessage_NoteOn_constructor(
    unittest.TestCase
):  # pylint: disable=invalid-name
    def test_NoteOn_constructor_string(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        object1 = NoteOn("C4", 0x64)
        self.assertEqual(object1.note, 60)
        self.assertEqual(object1.velocity, 0x64)

        object2 = NoteOn("C3", 0x7F)
        self.assertEqual(object2.note, 48)
        self.assertEqual(object2.velocity, 0x7F)

        object3 = NoteOn("C#4", 0x00)
        self.assertEqual(object3.note, 61)
        self.assertEqual(object3.velocity, 0)

    def test_NoteOn_constructor_valueerror1(self):  # pylint: disable=invalid-name
        with self.assertRaises(ValueError):
            NoteOn(60, 0x80)  # pylint is happier if return value not stored

    def test_NoteOn_constructor_valueerror2(self):  # pylint: disable=invalid-name
        with self.assertRaises(ValueError):
            NoteOn(-1, 0x7F)

    def test_NoteOn_constructor_valueerror3(self):  # pylint: disable=invalid-name
        with self.assertRaises(ValueError):
            NoteOn(128, 0x7F)

    def test_NoteOn_constructor_upperrange1(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        object1 = NoteOn("G9", 0x7F)
        self.assertEqual(object1.note, 127)
        self.assertEqual(object1.velocity, 0x7F)

    def test_NoteOn_constructor_upperrange2(self):  # pylint: disable=invalid-name
        with self.assertRaises(ValueError):
            NoteOn("G#9", 0x7F)  # just above max note

    def test_NoteOn_constructor_bogusstring(self):  # pylint: disable=invalid-name
        with self.assertRaises(ValueError):
            NoteOn("CC4", 0x7F)


class Test_MIDIMessage_NoteOff_constructor(
    unittest.TestCase
):  # pylint: disable=invalid-name
    # mostly cut and paste from NoteOn above
    def test_NoteOff_constructor_string(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        object1 = NoteOff("C4", 0x64)
        self.assertEqual(object1.note, 60)
        self.assertEqual(object1.velocity, 0x64)

        object2 = NoteOff("C3", 0x7F)
        self.assertEqual(object2.note, 48)
        self.assertEqual(object2.velocity, 0x7F)

        object3 = NoteOff("C#4", 0x00)
        self.assertEqual(object3.note, 61)
        self.assertEqual(object3.velocity, 0)

        object4 = NoteOff("C#4")  # velocity defaults to 0
        self.assertEqual(object4.note, 61)
        self.assertEqual(object4.velocity, 0)

    def test_NoteOff_constructor_valueerror1(self):  # pylint: disable=invalid-name
        with self.assertRaises(ValueError):
            NoteOff(60, 0x80)

    def test_NoteOff_constructor_valueerror2(self):  # pylint: disable=invalid-name
        with self.assertRaises(ValueError):
            NoteOff(-1, 0x7F)

    def test_NoteOff_constructor_valueerror3(self):  # pylint: disable=invalid-name
        with self.assertRaises(ValueError):
            NoteOff(128, 0x7F)

    def test_NoteOff_constructor_upperrange1(self):  # pylint: disable=invalid-name
        # pylint: enable=invalid-name
        object1 = NoteOff("G9", 0x7F)
        self.assertEqual(object1.note, 127)
        self.assertEqual(object1.velocity, 0x7F)

    def test_NoteOff_constructor_upperrange2(self):  # pylint: disable=invalid-name
        with self.assertRaises(ValueError):
            NoteOff("G#9", 0x7F)  # just above max note

    def test_NoteOff_constructor_bogusstring(self):  # pylint: disable=invalid-name
        with self.assertRaises(ValueError):
            NoteOff("CC4", 0x7F)


if __name__ == "__main__":
    unittest.main(verbosity=verbose)



# SPDX-FileCopyrightText: 2019 Kevin J. Walters for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""
`adafruit_midi.pitch_bend`
================================================================================

Pitch Bend Change MIDI message.


* Author(s): Kevin J. Walters

Implementation Notes
--------------------

"""

from .midi_message import MIDIMessage

__version__ = "0.0.0-auto.0"
__repo__ = "https://github.com/adafruit/Adafruit_CircuitPython_MIDI.git"


class PitchBend(MIDIMessage):
    """Pitch Bend Change MIDI message.

    :param int pitch_bend: A 14bit unsigned int representing the degree of
        bend from 0 through 8192 (midpoint, no bend) to 16383.
    """

    _STATUS = 0xE0
    _STATUSMASK = 0xF0
    LENGTH = 3

    def __init__(self, pitch_bend, *, channel=None):
        self.pitch_bend = pitch_bend
        super().__init__(channel=channel)
        if not 0 <= self.pitch_bend <= 16383:
            self._raise_valueerror_oor()

    def __bytes__(self):
        return bytes(
            [
                self._STATUS | (self.channel & self.CHANNELMASK),
                self.pitch_bend & 0x7F,
                (self.pitch_bend >> 7) & 0x7F,
            ]
        )

    @classmethod
    def from_bytes(cls, msg_bytes):
        return cls(
            msg_bytes[2] << 7 | msg_bytes[1], channel=msg_bytes[0] & cls.CHANNELMASK
        )


PitchBend.register_message_type()



# SPDX-FileCopyrightText: 2021 ladyada for Adafruit Industries
# SPDX-License-Identifier: MIT

# midi_inoutdemo - demonstrates receiving and sending MIDI events

import usb_midi
import adafruit_midi

# TimingClock is worth importing first if present as it
# will make parsing more efficient for this high frequency event
# Only importing what is used will save a little bit of memory

# pylint: disable=unused-import
from adafruit_midi.timing_clock import TimingClock
from adafruit_midi.channel_pressure import ChannelPressure
from adafruit_midi.control_change import ControlChange
from adafruit_midi.note_off import NoteOff
from adafruit_midi.note_on import NoteOn
from adafruit_midi.pitch_bend import PitchBend
from adafruit_midi.polyphonic_key_pressure import PolyphonicKeyPressure
from adafruit_midi.program_change import ProgramChange
from adafruit_midi.start import Start
from adafruit_midi.stop import Stop
from adafruit_midi.system_exclusive import SystemExclusive

from adafruit_midi.midi_message import MIDIUnknownEvent

midi = adafruit_midi.MIDI(
    midi_in=usb_midi.ports[0],
    midi_out=usb_midi.ports[1],
    in_channel=(1, 2, 3),
    out_channel=0,
)

print("Midi Demo in and out")

# Convert channel numbers at the presentation layer to the ones musicians use
print("Default output channel:", midi.out_channel + 1)
print("Listening on input channels:", tuple(c + 1 for c in midi.in_channel))

major_chord = [0, 4, 7]
while True:
    while True:
        msg_in = midi.receive()  # non-blocking read
        # For a Note On or Note Off play a major chord
        # For any other known event just forward it
        if isinstance(msg_in, NoteOn) and msg_in.velocity != 0:
            print(
                "Playing major chord with root",
                msg_in.note,
                "from channel",
                msg_in.channel + 1,
            )
            for offset in major_chord:
                new_note = msg_in.note + offset
                if 0 <= new_note <= 127:
                    midi.send(NoteOn(new_note, msg_in.velocity))

        elif (
            isinstance(msg_in, NoteOff)
            or isinstance(msg_in, NoteOn)
            and msg_in.velocity == 0
        ):
            for offset in major_chord:
                new_note = msg_in.note + offset
                if 0 <= new_note <= 127:
                    midi.send(NoteOff(new_note, 0x00))

        elif isinstance(msg_in, MIDIUnknownEvent):
            # Message are only known if they are imported
            print("Unknown MIDI event status ", msg_in.status)

        elif msg_in is not None:
            midi.send(msg_in)



# SPDX-FileCopyrightText: 2021 ladyada for Adafruit Industries
# SPDX-License-Identifier: MIT

# Check memory usage

# pylint: disable=multiple-statements,unused-import,wrong-import-position,no-member

# The disable for no-member should not really be required
# probably a difference between Python 3 module and micropython
#
# E:  8,21: Module 'gc' has no 'mem_free' member (no-member)

import time
import random
import gc

gc.collect()
print(gc.mem_free())
import usb_midi

gc.collect()
print(gc.mem_free())
import adafruit_midi

gc.collect()
print(gc.mem_free())

# Full monty
from adafruit_midi.channel_pressure import ChannelPressure

gc.collect()
print(gc.mem_free())
from adafruit_midi.control_change import ControlChange

gc.collect()
print(gc.mem_free())
from adafruit_midi.note_off import NoteOff

gc.collect()
print(gc.mem_free())
from adafruit_midi.note_on import NoteOn

gc.collect()
print(gc.mem_free())
from adafruit_midi.pitch_bend import PitchBend

gc.collect()
print(gc.mem_free())
from adafruit_midi.polyphonic_key_pressure import PolyphonicKeyPressure

gc.collect()
print(gc.mem_free())
from adafruit_midi.program_change import ProgramChange

gc.collect()
print(gc.mem_free())
from adafruit_midi.start import Start

gc.collect()
print(gc.mem_free())
from adafruit_midi.stop import Stop

gc.collect()
print(gc.mem_free())
from adafruit_midi.system_exclusive import SystemExclusive

gc.collect()
print(gc.mem_free())
from adafruit_midi.timing_clock import TimingClock

gc.collect()
print(gc.mem_free())

midi = adafruit_midi.MIDI(
    midi_in=usb_midi.ports[0], midi_out=usb_midi.ports[1], in_channel=0, out_channel=0
)

gc.collect()
print(gc.mem_free())



# SPDX-FileCopyrightText: 2019 Kevin J. Walters for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""
`adafruit_midi.start`
================================================================================

Start MIDI message.


* Author(s): Kevin J. Walters

Implementation Notes
--------------------

"""

from .midi_message import MIDIMessage

__version__ = "0.0.0-auto.0"
__repo__ = "https://github.com/adafruit/Adafruit_CircuitPython_MIDI.git"


class Start(MIDIMessage):
    """Start MIDI message."""

    _STATUS = 0xFA
    _STATUSMASK = 0xFF
    LENGTH = 1


Start.register_message_type()
